﻿/*
 * Example provided by Richard Fung, University of Calgary, October 2010
 * 
 * Phidget slider connected at index 0  --> change the threshold
 * Servo motor connected at index 0     --> indicates system's confidence in marker recognition
 * 
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using jp.nyatla.nyartoolkit.cs;
using jp.nyatla.nyartoolkit.cs.core;
using jp.nyatla.nyartoolkit.cs.detector;
using NyARToolkitCSUtils.Capture;
using NyARToolkitCSUtils.NyAR;


namespace WpfWebcam
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, CaptureListener
    {
        private const String AR_CODE_FILE = "../../data/patt.kanji";
        private const String AR_CAMERA_FILE = "../../data/camera_para.dat";

        private CaptureDevice m_cap;
        private NyARDetectMarker m_ar;

        private const double dpiX = 96.0;
        private const double dpiY = 96.0;

        private const int cameraResX = 640;
        private const int cameraResY = 480;

        private int m_threshold = 100;

        private NyARRgbRaster m_raster;
        // private DsBGRX32Raster m_raster; // use without AForge.Imaging

        private Phidgets.InterfaceKit m_kit;
        private Phidgets.Servo m_servo;


        public MainWindow()
        {
            InitializeComponent();


            //計算モードの設定
            //キャプチャを作る
            /**************************************************
            このコードは、0番目（一番初めに見つかったキャプチャデバイス）
            を使用するようにされています。
            複数のキャプチャデバイスを持つシステムの場合、うまく動作しないかもしれません。
            n番目のデバイスを使いたいときには、CaptureDevice cap=cl[0];←ここの0を変えてください。
            手動で選択させる方法は、SimpleLiteDirect3Dを参考にしてください。
            **************************************************/
            /*
             * This code here queries the list of DirectX webcams available on the sytem
             * and chooses the webcam it finds at index 0.
             */
            CaptureDeviceList cl = new CaptureDeviceList();
            m_cap = cl[0];
            m_cap.SetCaptureListener(this);
            m_cap.PrepareCapture(cameraResX, cameraResY, 15); // 800x600 resolution, 30 fps
            
            

            //ARの設定
            //AR用カメラパラメタファイルをロード
            // AR camera parameters file to read
            NyARParam ap = new NyARParam();
            ap.loadARParamFromFile(AR_CAMERA_FILE);
            ap.changeScreenSize(cameraResX, cameraResY);

            // create a raster to save the image from AForge.Imaging and thresholding (24 bpp image returned)
            this.m_raster = new NyARRgbRaster(m_cap.video_width, m_cap.video_height, NyARBufferType.BYTE1D_R8G8B8_24, false);

            // use this raster if taking the image directly without AForge.Imaging
            // this.m_raster = new DsBGRX32Raster(m_cap.video_width, m_cap.video_height, m_cap.video_width * 32 / 8); 

            //AR用のパターンコードを読み出し	
            // AR's pattern to detect from the webcam
            NyARCode code = new NyARCode(16, 16); // 16 pixels to detect within black box
            code.loadARPattFromFile(AR_CODE_FILE);
            this.m_ar = new NyARDetectMarker(ap, new NyARCode[]{code}, new double[]{80.0}, 1,NyARBufferType.BYTE1D_B8G8R8_24);
            this.m_ar.setContinueMode(false);

            // phidgets
            m_kit = new Phidgets.InterfaceKit();
            m_kit.open();

            m_servo = new Phidgets.Servo();
            m_servo.open();
        }

        void m_kit_SensorChange(object sender, Phidgets.Events.SensorChangeEventArgs e)
        {
            m_threshold = (int)((e.Value / 1000.0) * 255.0);

            Dispatcher.BeginInvoke(  new Action( delegate()  
                {
                    this.Title = "Threshold " + m_threshold;
                }));
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {

            // start the caputre
            this.m_cap.StartCapture();

            // start the phidgets
            m_kit.waitForAttachment(3000);
            m_servo.waitForAttachment(3000);

            // start listening
            m_kit.SensorChange += new Phidgets.Events.SensorChangeEventHandler(m_kit_SensorChange);

            // setup initial settings
            m_threshold = (int)((m_kit.sensors[0].Value / 1000.0) * 255.0);


        }


        private void Window_Unloaded(object sender, RoutedEventArgs e)
        {
            m_kit.SensorChange -= new Phidgets.Events.SensorChangeEventHandler(m_kit_SensorChange);
            m_cap.StopCapture();
            m_cap.Dump();
        }

        void CaptureListener.OnBuffer(CaptureDevice i_sender, double i_sample_time, IntPtr i_buffer, int i_buffer_len)
        {
            // calculate size of the frame bitmap
            int w = i_sender.video_width;
            int h = i_sender.video_height;
            int s = w * (i_sender.video_bit_count / 8); // stride            
            
            
            // thresholding algorithm
            AForge.Imaging.Filters.FiltersSequence seq = new AForge.Imaging.Filters.FiltersSequence();

            // order here is important, pipe and filters design pattern
            seq.Add(new AForge.Imaging.Filters.Grayscale(0.2125, 0.7154, 0.0721));
            //seq.Add(AForge.Imaging.Filters.Grayscale.CommonAlgorithms.BT709);
            seq.Add(new AForge.Imaging.Filters.Threshold(m_threshold));
            //seq.Add(new AForge.Imaging.Filters.OtsuThreshold());
            seq.Add(new AForge.Imaging.Filters.GrayscaleToRGB()); // 24 bit image returned

            // run the threshold algorithm
            AForge.Imaging.UnmanagedImage srcImg = new AForge.Imaging.UnmanagedImage(i_buffer, w,h, s, System.Drawing.Imaging.PixelFormat.Format32bppRgb);
            AForge.Imaging.UnmanagedImage outputImg = seq.Apply(srcImg);

            //// load a RGB buffer into the NyAR format, which we copy from AForge.Imaging
            byte[] destArr = new byte[outputImg.Stride * outputImg.Height];
            System.Runtime.InteropServices.Marshal.Copy(outputImg.ImageData, destArr, 0, outputImg.Stride * outputImg.Height);
            this.m_raster.wrapBuffer(destArr);

            // // use this code if you want to skip thresholding --> it won't work well!!!
            // // and switch the m_raster to DsBGRX32Raster
            // byte[] destArr = new byte[i_buffer_len];
            // System.Runtime.InteropServices.Marshal.Copy(i_buffer, destArr, 0, i_buffer_len);
            // this.m_raster.wrapBuffer(destArr);
            
            
            try
            {
                // detectMarkerLite requires BGR image from DirectX
                // threshold 0-255
                int detectedMkrs = this.m_ar.detectMarkerLite(this.m_raster, m_threshold);

                // save the result of the detection
                NyARSquare square = null;
                if (detectedMkrs > 0)
                {
                    // vertices of the square are returned
                    NyARTransMatResult transMat = new NyARTransMatResult();
                    NyARDoublePoint2d[] points = m_ar.getCorners(0); // RichF added this method
                    square = new NyARSquare();
                    square.sqvertex = points;

                }
            

                // make the image on the window on the UI thread
                Dispatcher.BeginInvoke(  new Action( delegate()  
                {
                    // create the output bitmap
                    TransformedBitmap b = new TransformedBitmap();
                    b.BeginInit();

                    // load the bitmap data from raw pixels
                    // dpiX, dpiY:  higher creates smaller picture on the window
                    // WebPalette:  not sure why we use it
                    // Bgr32:       matches DsBGRX32Raster from DirectX

                    b.Source = BitmapSource.Create(w, h, dpiX, dpiY, PixelFormats.Bgr32, BitmapPalettes.WebPalette, i_buffer, i_buffer_len, s);

                    // use this line for false colour image
                    //b.Source = BitmapSource.Create(w, h, dpiX, dpiY, PixelFormats.Bgr24, BitmapPalettes.WebPalette, outputImg.ImageData, outputImg.Stride * outputImg.Height, outputImg.Stride);

                    // flip the image because raw bitmap data is upside down
                    b.SetValue(TransformedBitmap.TransformProperty, new ScaleTransform(1, -1));

                    // finalize the loading
                    b.EndInit();

                    // draw the image on the form
                    image1.SetValue(Image.SourceProperty, b);

                    if (square!=null)
                    {
                        // just outline the corners of the square
                        // remember, the recognizer performed operations upside down
                        recognizedTag.Points = new PointCollection(new Point[] { new Point(square.sqvertex[0].x, cameraResY - square.sqvertex[0].y),
                                new Point(square.sqvertex[1].x, cameraResY - square.sqvertex[1].y),
                                new Point(square.sqvertex[2].x, cameraResY - square.sqvertex[2].y),
                                new Point(square.sqvertex[3].x, cameraResY - square.sqvertex[3].y)});

                        recognizedTag.Visibility = System.Windows.Visibility.Visible;
                        lock (m_servo)
                        {
                            // boundary check
                            double conf = m_ar.getConfidence(0);
                            if (conf < 0) conf = 0;
                            if (conf > 1) conf = 1;

                            // compute
                            m_servo.servos[0].Position = (m_servo.servos[0].PositionMax - m_servo.servos[0].PositionMin) * conf + m_servo.servos[0].PositionMin;
                        }
                    }
                    else
                    {
                        recognizedTag.Visibility = System.Windows.Visibility.Hidden;
                        lock (m_servo)
                        {
                            m_servo.servos[0].Position = m_servo.servos[0].PositionMin;
                        }
                    }
                }), null);

            }
            catch
            {

            }

        }

        private void Window_KeyDown(object sender, KeyEventArgs e)
        {
            // press arrow keys to change threshold
            if (e.Key == Key.Up)
            {
                m_threshold--;
            }
            if (e.Key == Key.Down)
            {
                m_threshold++;
            }

            Title = "Threshold " + m_threshold;
        }


    }
}
